{ File   : MemSearch.pas
  Author : Deniz Oezmen
  Created: 2010-10-16, based on Utils.pas created 2008-08-21
  Changed: 2010-10-24

  Memory search utilities.
}
unit MemSearch;

interface

type
  TSearchMethod = (SearchMethod_RabinKarp, SearchMethod_BoyerMoore,
    SearchMethod_KnuthMorrisPratt);

function SearchMem(const SearchMethod: TSearchMethod;
  Pattern, Data: Pointer; PatternSize, DataSize: Integer): Integer;
function SearchMemReverse(const SearchMethod: TSearchMethod;
  Pattern, Data: Pointer; PatternSize, DataSize: Integer): Integer;
procedure SearchMaxPrefix(const SearchMethod: TSearchMethod;
  Pattern, Data: Pointer; PatternSize, DataSize: Integer;
  out MatchPos, MatchLen: Integer);

implementation

uses
  SysUtils;

// Rabin-Karp search algorithm as presented in Sedgewick, Algorithms, Chapter 19
function RabinKarp(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer): Integer;
const
  q = 8388593; // largest prime such that q * 256 <= 2^31 - 1
type
  PByte = ^Byte;
var
  i,
  N, M      : Integer;
  h1, h2,
  dM        : Longword;
  PatternPtr,
  DataPtr   : PByte;
begin
  Result := -1;

  N := DataSize;
  M := PatternSize;

  dM := 1;
  for i := 1 to M - 1 do
    dM := (dM shl 8) mod q;

  h1 := 0;
  PatternPtr := Pattern;
  for i := 1 to M do
  begin
    h1 := (h1 shl 8 + PatternPtr^) mod q;
    Inc(PatternPtr)
  end;

  h2 := 0;
  DataPtr := Data;
  for i := 1 to M do
  begin
    h2 := (h2 shl 8 + DataPtr^) mod q;
    Inc(DataPtr)
  end;

  i := 0;
  DataPtr := Data;
  PatternPtr := Pattern;
  repeat
    if (h1 = h2) and (CompareMem(DataPtr, PatternPtr, M)) then
    begin
      Result := i;
      Break
    end;

    h2 := (h2 + q shl 8 - DataPtr^ * dM) mod q;
    h2 := (h2 shl 8 + PByte(Integer(DataPtr) + M)^) mod q;

    Inc(i);
    Inc(DataPtr)
  until i > N - M
end;

// same algorithm as above, but searches through Data in reverse order,
// effectively returning the *last* occurrence of the search pattern
function RabinKarpReverse(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer): Integer;
const
  q = 8388593; // largest prime such that q * 256 <= 2^31 - 1
type
  PByte = ^Byte;
var
  i,
  N, M      : Integer;
  h1, h2,
  dM        : Longword;
  PatternPtr,
  DataPtr   : PByte;
begin
  Result := -1;

  N := DataSize;
  M := PatternSize;

  dM := 1;
  for i := 1 to M - 1 do
    dM := (dM shl 8) mod q;

  h1 := 0;
  PatternPtr := PByte(Integer(Pattern) + M - 1);
  for i := 1 to M do
  begin
    h1 := (h1 shl 8 + PatternPtr^) mod q;
    Dec(PatternPtr)
  end;

  h2 := 0;
  DataPtr := PByte(Integer(Data) + N - 1);
  for i := 1 to M do
  begin
    h2 := (h2 shl 8 + DataPtr^) mod q;
    Dec(DataPtr)
  end;

  i := N - M;
  DataPtr := PByte(Integer(Data) + N - 1);
  PatternPtr := Pattern;
  repeat
    if (h1 = h2) and (CompareMem(PByte(Integer(DataPtr) - M + 1), PatternPtr,
      M)) then
    begin
      Result := i;
      Break
    end;

    h2 := (h2 + q shl 8 - DataPtr^ * dM) mod q;
    h2 := (h2 shl 8 + PByte(Integer(DataPtr) - M)^) mod q;

    Dec(i);
    Dec(DataPtr)
  until i < 0
end;

// algorithm extended to find the maximum length prefix of Pattern within Data 
procedure RabinKarpPrefix(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer; out MatchPos, MatchLen: Integer);
const
  q = 8388593; // largest prime such that q * 256 <= 2^31 - 1
type
  PByte = ^Byte;
var
  i, j,
  N, M      : Integer;
  h1, h2,
  dM        : Longword;
  PatternPtr,
  DataPtr   : PByte;
  PatternHsh,
  dMs,
  h2s       : array of Longword;
begin
  MatchPos := -1;
  MatchLen := 0;

  N := DataSize;
  M := PatternSize;

  if (N <= 0) or (M <= 0) then
    Exit;

  if M > N then
    M := N;

  SetLength(dMs, M);
  dM := 1;
  dMs[0] := dM;
  for i := 1 to M - 1 do
  begin
    dM := (dM shl 8) mod q;
    dMs[i] := dM
  end;

  SetLength(PatternHsh, M);
  h1 := 0;
  PatternPtr := Pattern;
  for i := 0 to M - 1 do
  begin
    h1 := (h1 shl 8 + PatternPtr^) mod q;
    Inc(PatternPtr);

    PatternHsh[i] := h1
  end;

  SetLength(h2s, M);
  h2 := 0;
  DataPtr := Data;
  for i := 0 to M - 1 do
  begin
    h2 := (h2 shl 8 + DataPtr^) mod q;
    h2s[i] := h2;
    Inc(DataPtr)
  end;

  i := 0;
  DataPtr := Data;
  PatternPtr := Pattern;
  repeat
    if i + M - 1 >= N then
      M := N - i;

    if MatchLen >= M then
      Break;

    for j := MatchLen to M - 1 do
      if PatternHsh[j] = h2s[j] then
        if CompareMem(DataPtr, PatternPtr, j + 1) then
        begin
          MatchPos := i;
          MatchLen := j + 1
        end;

    for j := MatchLen to M - 1 do
    begin
      h2 := h2s[j];
      h2 := (h2 + q shl 8 - DataPtr^ * dMs[j]) mod q;
      h2 := (h2 shl 8 + PByte(Integer(DataPtr) + j + 1)^) mod q;
      h2s[j] := h2
    end;

    Inc(i);
    Inc(DataPtr)
  until i > N - MatchLen
end;

function BoyerMoore(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer): Integer;
var
  i, j,
  N, M    : Integer;
  DataByte: Byte;
  Skip    : array[0..255] of Integer;
  Next    : array of Integer;

  procedure InitSkip;
  var
    i: Integer;
  begin
    for i := 0 to 255 do
      Skip[i] := M;

    for i := 0 to M - 2 do
      Skip[PByte(Integer(Pattern) + i)^] := M - i - 1
  end;

  // Boyer-Moore "Next" implementation adapted from
  // http://www-igm.univ-mlv.fr/~lecroq/string/node14.html
  procedure InitNext;
  var
    i, j,
    LastI : Integer;
    Suffix: array of Integer;
  begin
    SetLength(Suffix, M);
    SetLength(Next, M);

    Suffix[M - 1] := M;
		j := M - 1;
    LastI := 0;

    for i := M - 2 downto 0 do
    begin
      if (i > j) and (Suffix[i + M - 1 - LastI] < i - j) then
        Suffix[i] := Suffix[i + M - 1 - LastI]
      else
      begin
        if i < j then
          j := i;

        while (j >= 0) and (PByte(Integer(Pattern) + j)^ =
          PByte(Integer(Pattern) + j + M - 1 - i)^) do
          Dec(j);

        Suffix[i] := i - j;
        LastI := i
      end
    end;

    for i := 0 to M - 1 do
      Next[i] := M;

    j := 0;
    for i := M - 1 downto -1 do
      if (i = -1) or (Suffix[i] = i + 1) then
        while j < M - 1 - i do
        begin
          if Next[j] = M then
            Next[j] := M - 1 - i;
          Inc(j)
        end;

    for i := 0 to M - 2 do
      Next[M - 1 - Suffix[i]] := M - 1 - i
  end;

begin
  Result := - 1;

  N := DataSize;
  M := PatternSize;
  if (M > N) or (M = 0) or (N = 0) then
    Exit;

  InitSkip;
  InitNext;

  i := M - 1;

  while i < N do
  begin
    j := 0;

    while j < M do
    begin
      DataByte := PByte(Integer(Data) + i)^;

      if PByte(Integer(Data) + i - j)^ =
        PByte(Integer(Pattern) + M - j - 1)^ then
        Inc(j)
      else
      begin
        if Next[M - j - 1] > Skip[DataByte] then
          Inc(i, Next[M - j - 1])
        else
          Inc(i, Skip[DataByte]);

        Break
      end
    end;

    if j >= M then
      Break
  end;

  if i >= N then
    Result := -1
  else
    Result := i - M + 1;

  SetLength(Next, 0)
end;

function BoyerMooreReverse(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer): Integer;
var
  i, j,
  N, M    : Integer;
  DataByte: Byte;
  Skip    : array[0..255] of Integer;
  Next    : array of Integer;

  procedure InitSkip;
  var
    i: Integer;
  begin
    for i := 0 to 255 do
      Skip[i] := M;

    for i := M - 1 downto 1 do
      Skip[PByte(Integer(Pattern) + i)^] := i
  end;

  // Boyer-Moore "Next" implementation adapted from
  // http://www-igm.univ-mlv.fr/~lecroq/string/node14.html
  procedure InitNext;
  var
    i, j,
    LastI : Integer;
    Suffix: array of Integer;
  begin
    SetLength(Suffix, M);
    SetLength(Next, M);

    Suffix[M - 1] := M;
		j := M - 1;
    LastI := 0;

    for i := M - 2 downto 0 do
    begin
      if (i > j) and (Suffix[i + M - 1 - LastI] < i - j) then
        Suffix[i] := Suffix[i + M - 1 - LastI]
      else
      begin
        if i < j then
          j := i;

        while (j >= 0) and (PByte(Integer(Pattern) + M - 1 - j)^ =
          PByte(Integer(Pattern) + i - j)^) do
          Dec(j);

        Suffix[i] := i - j;
        LastI := i
      end
    end;

    for i := 0 to M - 1 do
      Next[i] := M;

    j := 0;
    for i := M - 1 downto -1 do
      if (i = -1) or (Suffix[i] = i + 1) then
        while j < M - 1 - i do
        begin
          if Next[j] = M then
            Next[j] := M - 1 - i;
          Inc(j)
        end;

    for i := 0 to M - 2 do
      Next[M - 1 - Suffix[i]] := M - 1 - i
  end;

begin
  Result := - 1;

  N := DataSize;
  M := PatternSize;
  if (M > N) or (M = 0) or (N = 0) then
    Exit;

  InitSkip;
  InitNext;

  i := N - M;

  while i >= 0 do
  begin
    j := 0;

    while j < M do
    begin
      DataByte := PByte(Integer(Data) + i)^;

      if PByte(Integer(Data) + i + j)^ = PByte(Integer(Pattern) + j)^ then
        Inc(j)
      else
      begin
        if Next[M - j - 1] > Skip[DataByte] then
          Dec(i, Next[M - j - 1])
        else
          Dec(i, Skip[DataByte]);
        Break
      end
    end;

    if j >= M then
      Break
  end;

  if i < 0 then
    Result := -1
  else
    Result := i;

  SetLength(Next, 0)
end;

procedure BoyerMoorePrefix(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer; out MatchPos, MatchLen: Integer);
var
  Min, Max,
  PrefixLen,
  PrefixPos: Integer;
begin
  MatchPos := -1;
  MatchLen := 0;

  // do a sort of binary search over the prefix length
  Min := 1;
  Max := PatternSize;

  if Max < Min then
    Exit;

  repeat
    PrefixLen := Min + (Max - Min) div 2;

    PrefixPos := BoyerMoore(Pattern, Data, PrefixLen, DataSize);
    if PrefixPos < 0 then
      Max := PrefixLen - 1
    else
    begin
      MatchPos := PrefixPos;
      MatchLen := PrefixLen;
      Min := PrefixLen + 1
    end
  until Min > Max
end;

function KnuthMorrisPratt(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer): Integer;
var
  i, j,
  N, M    : Integer;
  DataByte: Byte;
  Next    : array of Integer;

  procedure InitNext;
  var
    i, j       : Integer;
    PatternByte: Byte;
  begin
    SetLength(Next, M + 1);

    i := 0;
    j := -1;

    Next[i] := j;

    while i < M do
    begin
      PatternByte := PByte(Integer(Pattern) + i)^;

      while (j >= 0) and (PByte(Integer(Pattern) + j)^ <> PatternByte) do
        j := Next[j];

      i := i + 1;
      j := j + 1;

      Next[i] := j
    end
  end;

begin
  Result := - 1;

  N := DataSize;
  M := PatternSize;
  if (M > N) or (M = 0) or (N = 0) then
    Exit;

  InitNext;

  i := 0;
  j := 0;

  while i < N do
  begin
    DataByte := PByte(Integer(Data) + i)^;

    while (j >= 0) and (DataByte <> PByte(Integer(Pattern) + j)^) do
      j := Next[j];

    i := i + 1;
    j := j + 1;

    if j = M then
    begin
      Result := i - M;
      Break // for further searches set j := Next[j]
    end
  end
end;

function KnuthMorrisPrattReverse(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer): Integer;
var
  i, j,
  N, M    : Integer;
  DataByte: Byte;
  Next    : array of Integer;

  procedure InitNext;
  var
    i, j       : Integer;
    PatternByte: Byte;
  begin
    SetLength(Next, M + 1);

    i := M - 1;
    j := M;

    Next[i] := j;

    while i >= 0 do
    begin
      PatternByte := PByte(Integer(Pattern) + i)^;

      while (j < M) and (PByte(Integer(Pattern) + j)^ <> PatternByte) do
        j := Next[j];

      i := i - 1;
      j := j - 1;

      Next[i] := j
    end
  end;

begin
  Result := - 1;

  N := DataSize;
  M := PatternSize;
  if (M > N) or (M = 0) or (N = 0) then
    Exit;

  InitNext;

  i := N - 1;
  j := M - 1;

  while i >= 0 do
  begin
    DataByte := PByte(Integer(Data) + i)^;

    while (j < M) and (DataByte <> PByte(Integer(Pattern) + j)^) do
      j := Next[j];

    i := i - 1;
    j := j - 1;

    if j < 0 then
    begin
      Result := i + 1;
      Break
    end
  end
end;

procedure KnuthMorrisPrattPrefix(Pattern, Data: Pointer;
  PatternSize, DataSize: Integer; out MatchPos, MatchLen: Integer);
var
  i, j,
  N, M    : Integer;
  DataByte: Byte;
  Next    : array of Integer;

  procedure InitNext;
  var
    i, j       : Integer;
    PatternByte: Byte;
  begin
    SetLength(Next, M + 1);

    i := 0;
    j := -1;

    Next[i] := j;

    while i < M do
    begin
      PatternByte := PByte(Integer(Pattern) + i)^;

      while (j >= 0) and (PByte(Integer(Pattern) + j)^ <> PatternByte) do
        j := Next[j];

      i := i + 1;
      j := j + 1;

      Next[i] := j
    end
  end;

begin
  MatchPos := -1;
  MatchLen := 0;

  N := DataSize;
  M := PatternSize;
  if (M = 0) or (N = 0) then
    Exit;

  InitNext;

  i := 0;
  j := 0;

  while i < N do
  begin
    DataByte := PByte(Integer(Data) + i)^;

    while (j >= 0) and (DataByte <> PByte(Integer(Pattern) + j)^) do
      j := Next[j];

    i := i + 1;
    j := j + 1;

    if j > MatchLen then
    begin
      MatchPos := i - j;
      MatchLen := j
    end;

    if j = M then
      Break
  end
end;

function SearchMem(const SearchMethod: TSearchMethod;
  Pattern, Data: Pointer; PatternSize, DataSize: Integer): Integer;
begin
  if SearchMethod = SearchMethod_RabinKarp then
    Result := RabinKarp(Pattern, Data, PatternSize, DataSize)
  else if SearchMethod = SearchMethod_BoyerMoore then
    Result := BoyerMoore(Pattern, Data, PatternSize, DataSize)
  else if SearchMethod = SearchMethod_KnuthMorrisPratt then
    Result := KnuthMorrisPratt(Pattern, Data, PatternSize, DataSize)
  else
    Result := -1
end;

function SearchMemReverse(const SearchMethod: TSearchMethod;
  Pattern, Data: Pointer; PatternSize, DataSize: Integer): Integer;
begin
  if SearchMethod = SearchMethod_RabinKarp then
    Result := RabinKarpReverse(Pattern, Data, PatternSize, DataSize)
  else if SearchMethod = SearchMethod_BoyerMoore then
    Result := BoyerMooreReverse(Pattern, Data, PatternSize, DataSize)
  else if SearchMethod = SearchMethod_KnuthMorrisPratt then
    Result := KnuthMorrisPrattReverse(Pattern, Data, PatternSize, DataSize)
  else
    Result := -1
end;

procedure SearchMaxPrefix(const SearchMethod: TSearchMethod;
  Pattern, Data: Pointer; PatternSize, DataSize: Integer;
  out MatchPos, MatchLen: Integer);
begin
  if SearchMethod = SearchMethod_RabinKarp then
    RabinKarpPrefix(Pattern, Data, PatternSize, DataSize, MatchPos,
      MatchLen)
  else if SearchMethod = SearchMethod_BoyerMoore then
    BoyerMoorePrefix(Pattern, Data, PatternSize, DataSize, MatchPos,
      MatchLen)
  else if SearchMethod = SearchMethod_KnuthMorrisPratt then
    KnuthMorrisPrattPrefix(Pattern, Data, PatternSize, DataSize, MatchPos,
      MatchLen)
  else
    MatchPos := -1
end;

end.

